#include <argp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>
#include "comm.h"

#include "interrupt_trace.skel.h"
#include "schedule_trace.skel.h"

#include "hashmap.h"
#include "helpers.h"
#include "trace_helpers.h"
#include "syscall_helpers.h"
#include "shm_comm.h"

static volatile bool exiting = false;

static struct env {
	int self_tgid;
	bool enable_interrupt;
	bool enable_schedule;
	int sc_prev_tgid;
} env = {
	.enable_interrupt = false,
	.enable_schedule = false,
	.sc_prev_tgid = 0,
};


struct interrupt_trace_bpf *interrupt_skel = NULL;
struct schedule_trace_bpf *schedule_skel = NULL;


static int intmap_fd;
static int schedmap_fd;


static key_t shm_key;
static int shm_id;
static char* shm_start;
static char* shm_begin;



const char argp_program_doc[] ="Trace the lifecycle of  target process to get process key information.\n";

static const struct argp_option opts[] = {
	{ "all", 'a', NULL, 0, "Enable all function" },
	{ "interrupt", 'i', NULL, 0, "Enable interrupt function" },
	{ "schedule", 's', NULL, 0, "Enable schedule function" },
    { NULL, 'h', NULL, OPTION_HIDDEN, "show the help" },
	{},
};

static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
	switch (key) {
		case 'a':
				env.enable_interrupt = true;
				env.enable_schedule = true;
				break;
		case 'i':
				env.enable_interrupt = true;
                break;
		case 's':
				env.enable_schedule = true;
				break;
		case 'h':
				argp_state_help(state, stderr, ARGP_HELP_STD_HELP);
				break;
        default:
				return ARGP_ERR_UNKNOWN;
	}
	
	return 0;
}

static int print_schedule(void *ctx, void *data, unsigned long data_sz)
{
	while(GET_REMAINING_SPACE(shm_begin) < data_sz) {
		usleep(30);
		continue;
	}
	enum event_type* eventp = (enum event_type*)data;
	unsigned long *k_ip = NULL;
	unsigned long *u_ip = NULL;

	// int tgid = 0;
	if(need_stack(eventp)) {
		// get ip;
		int stack_fd = bpf_map__fd(schedule_skel->maps.stackmap);
		// tgid = *(int*)((char*)data + sizeof(enum event_type));
		int kern_stack_id = *(int*)((char*)data + sizeof(enum event_type) + sizeof(int));
		int user_stack_id = *(int*)((char*)data + sizeof(enum event_type) + 2*sizeof(int));
		k_ip = calloc(STACK_DEPTH, sizeof(*k_ip));
		u_ip = calloc(STACK_DEPTH, sizeof(*u_ip));
		bpf_map_lookup_elem(stack_fd, &kern_stack_id, k_ip);
		bpf_map_lookup_elem(stack_fd, &user_stack_id, u_ip);
	}
	int index = 0;
	int offset = 0;
	int i = 0, j = 0;
	unsigned rear = SHM_GET_REAR(shm_begin);
	for(i = 0; i < data_sz; i++) {
		offset = (rear + i) % SHM_SIZE;
		*(shm_start + offset) = *((char*)data + i);
	}
	rear = (rear + data_sz) % SHM_SIZE;
	if(need_stack(eventp)) {
		for(i = 0; i < STACK_DEPTH; i++) {
			for(j = 0; j < sizeof(*k_ip); j++) {
				index = i*sizeof(*k_ip) + j;
				offset = (rear + index) % SHM_SIZE;
				*(shm_start + offset) = *((char*)k_ip + index);
			}
		}
		rear = (rear + STACK_DEPTH*sizeof(*k_ip)) % SHM_SIZE;

		for(i = 0; i < STACK_DEPTH; i++) {
			for(j = 0; j < sizeof(*u_ip); j++) {
				index = i*sizeof(*u_ip) + j;
				offset = (rear + index) % SHM_SIZE;
				*(shm_start + offset) = *((char*)u_ip + index);
			}
		}
		rear = (rear + STACK_DEPTH*sizeof(*u_ip)) % SHM_SIZE;
	}
	SHM_GET_REAR(shm_begin) = rear;
	free(k_ip);
	free(u_ip);
	return 0;
}


static int print_interrupt(void *ctx, void *data, unsigned long data_sz)
{
	// enum event_type* eventp = (enum event_type*)data;
	while(GET_REMAINING_SPACE(shm_begin) < data_sz) {
		usleep(30);
		continue;
	}
	unsigned rear = SHM_GET_REAR(shm_begin);
	for(int i = 0; i < data_sz; i++) {
		int offset = (rear + i) % SHM_SIZE;
		*(shm_start + offset) = *((char*)data + i);
	}
	SHM_GET_REAR(shm_begin) = (rear + data_sz) % SHM_SIZE; 
	return 0;
}


static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
{
	return vfprintf(stderr, format, args);
}



static void sig_handler(int signo)
{
	exiting = true;
}

int main(int argc, char **argv)
{
	struct ring_buffer *interrupt_rb = NULL;
	struct bpf_map *interrupt_ctrl_map = NULL;
	struct ring_buffer *sched_rb = NULL;
	struct bpf_map *sched_ctrl_map = NULL;

	int key = 0;
	int err;
	static const struct argp argp = {
		.options = opts,
		.parser = parse_arg,
		.doc = argp_program_doc,
	};

	err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
	if (err)
		return err;
	
	env.self_tgid = getpid();

	libbpf_set_strict_mode(LIBBPF_STRICT_ALL);
	/* 设置libbpf错误和调试信息回调 */
	libbpf_set_print(libbpf_print_fn);

	signal(SIGINT, sig_handler);
	signal(SIGTERM, sig_handler);

    shm_key = getKey();
	if(shm_key == -1) {
        fprintf(stderr, "shm ftok error\n");
    }
	shm_id = createShm(shm_key);
	if(shm_id == -1) {
        fprintf(stderr, "shm shmget error\n");
    }
	shm_begin = processAssociation(shm_id);
    if((void*)shm_begin == (void*)-1) {
        fprintf(stderr, "shm attach error\n");
    }
    //将shm清空
	memset(shm_begin, 0, SHM_TOTAL_SIZE);
	shm_start = shm_begin + SHM_START_OFFSET;
	if(env.enable_interrupt){
		interrupt_skel = interrupt_trace_bpf__open();
		if(!interrupt_skel) {
			fprintf(stderr, "Failed to open BPF syscall skeleton\n");
			return 1;
		}
		
		err = interrupt_trace_bpf__load(interrupt_skel);
		if (err) {
			fprintf(stderr, "Failed to load and verify BPF syscall skeleton\n");
			goto cleanup;
		}

		err = common_pin_map(&interrupt_ctrl_map,interrupt_skel->obj,"interrupt_ctrl_map",interrupt_ctrl_path);
		if(err < 0){
			goto cleanup;
		}
		intmap_fd = bpf_map__fd(interrupt_ctrl_map);
		struct interrupt_ctrl init_value= {false,-1};
		err = bpf_map_update_elem(intmap_fd, &key, &init_value, 0);
		if(err < 0){
			fprintf(stderr, "Failed to update elem\n");
			goto cleanup;
		}

		err = interrupt_trace_bpf__attach(interrupt_skel);
		if (err) {
			fprintf(stderr, "Failed to attach BPF syscall skeleton\n");
			goto cleanup;
		}

		/* 设置环形缓冲区轮询 */
		//ring_buffer__new() API，允许在不使用额外选项数据结构下指定回调
		interrupt_rb = ring_buffer__new(bpf_map__fd(interrupt_skel->maps.interrupt_rb), print_interrupt, NULL, NULL);
		if (!interrupt_rb) {
			err = -1;
			fprintf(stderr, "Failed to create syscall ring buffer\n");
			goto cleanup;
		}
	}

	if(env.enable_schedule){
		schedule_skel = schedule_trace_bpf__open();
		if(!schedule_skel) {
			fprintf(stderr, "Failed to open BPF schedule skeleton\n");
			return 1;
		}

		bpf_map__set_value_size(schedule_skel->maps.stackmap,
					STACK_DEPTH * sizeof(unsigned long));
		bpf_map__set_max_entries(schedule_skel->maps.stackmap, 10240);
	// if (!probe_tp_btf("sched_switch"))
		// bpf_program__set_autoload(schedule_skel->progs.sched_switch, false);
	// else
	// 	bpf_program__set_autoload(schedule_skel->progs.sched_switch_raw, false);

		err = schedule_trace_bpf__load(schedule_skel);
		if (err) {
			fprintf(stderr, "Failed to load and verify BPF schedule skeleton\n");
			goto cleanup;
		}
		// 用户态获得map
		err = common_pin_map(&sched_ctrl_map,schedule_skel->obj,"sched_ctrl_map",sched_ctrl_path);
		if(err < 0){
			goto cleanup;
		}
		schedmap_fd = bpf_map__fd(sched_ctrl_map);
		struct sched_ctrl init_value= {false,-1};
		err = bpf_map_update_elem(schedmap_fd, &key, &init_value, 0);
		if(err < 0){
			fprintf(stderr, "Failed to update elem\n");
			goto cleanup;
		}

		err = schedule_trace_bpf__attach(schedule_skel);
		if (err) {
			fprintf(stderr, "Failed to attach BPF schedule skeleton\n");
			goto cleanup;
		}
		sched_rb = ring_buffer__new(bpf_map__fd(schedule_skel->maps.sched_rb), print_schedule, NULL, NULL);
		if (!sched_rb) {
			err = -1;
			fprintf(stderr, "Failed to create schedule ring buffer\n");
			goto cleanup;
		}
	}
	printf("start processing\n");
	/* 处理事件 */
	while (!exiting) {
		if(env.enable_interrupt){
			err = ring_buffer__poll(interrupt_rb, 0);
			/* Ctrl-C will cause -EINTR */
			if (err == -EINTR) {
				err = 0;
				break;
			}
			if (err < 0) {
				fprintf(stderr, "Error polling interrupt ring buffer: %d\n", err);
				break;
			}
		}

		if(env.enable_schedule){
			err = ring_buffer__poll(sched_rb, 0);
			/* Ctrl-C will cause -EINTR */
			if (err == -EINTR) {
				err = 0;
				break;
			}
			if (err < 0) {
				fprintf(stderr, "Error polling schedule ring buffer: %d\n", err);
				break;
			}
		}
	}

/* 卸载BPF程序 */
cleanup:
	if(deAssociation(shm_begin) == -1) {
		fprintf(stderr, "shm deattach error\n");
	}
	shmRelease(shm_id);
	if(env.enable_schedule){
		bpf_map__unpin(sched_ctrl_map, sched_ctrl_path);
		ring_buffer__free(sched_rb);
		schedule_trace_bpf__destroy(schedule_skel);
	}
	if(env.enable_interrupt){
		bpf_map__unpin(interrupt_ctrl_map, interrupt_ctrl_path);
		ring_buffer__free(interrupt_rb);
		interrupt_trace_bpf__destroy(interrupt_skel);
	}
	return err < 0 ? -err : 0;
}